#include "td_main.h"
#include <raylib.h>
#include <stdio.h>

typedef struct ParserState
{
  char *input;
  int position;
  char nextToken[256];
} ParserState;

void ParserStateSkipWhiteSpaces(ParserState *state)
{
  char *input = state->input;
  int pos = state->position;
  int skipped = 1;
  while (skipped)
  {
    skipped = 0;
    if (input[pos] == '-' && input[pos + 1] == '-')
    {
      skipped = 1;
      // skip comments
      while (input[pos] != 0 && input[pos] != '\n')
      {
        pos++;
      }
    }
  
    // skip white spaces and ignore colons
    while (input[pos] != 0 && (input[pos] <= ' ' || input[pos] == ':'))
    {
      skipped = 1;
      pos++;
    }

    // repeat until no more white spaces or comments
  }
  state->position = pos;
}

int ParserStateReadNextToken(ParserState *state)
{
  ParserStateSkipWhiteSpaces(state);

  int i = 0, pos = state->position;
  char *input = state->input;

  // read token
  while (input[pos] != 0 && input[pos] > ' ' && input[pos] != ':' && i < 256)
  {
    state->nextToken[i] = input[pos];
    pos++;
    i++;
  }
  state->position = pos;

  if (i == 0 || i == 256)
  {
    state->nextToken[0] = 0;
    return 0;
  }
  // terminate the token
  state->nextToken[i] = 0;
  return 1;
}

int ParserStateReadNextInt(ParserState *state, int *value)
{
  if (!ParserStateReadNextToken(state))
  {
    return 0;
  }
  // check if the token is a number
  int isSigned = state->nextToken[0] == '-' || state->nextToken[0] == '+';
  for (int i = isSigned; state->nextToken[i] != 0; i++)
  {
    if (state->nextToken[i] < '0' || state->nextToken[i] > '9')
    {
      return 0;
    }
  }
  *value = TextToInteger(state->nextToken);
  return 1;
}

typedef struct ParsedGameData
{
  char *parseError;
  Level levels[32];
  int lastLevelIndex;
  TowerTypeConfig towerTypes[TOWER_TYPE_COUNT];
  EnemyClassConfig enemyClasses[8];
} ParsedGameData;

int ParseGameDataLevelSection(ParserState *state, ParsedGameData *gameData)
{
  int levelId;
  if (!ParserStateReadNextInt(state, &levelId))
  {
    gameData->parseError = "Failed to read level id";
    return 0;
  }
  
  if (levelId < 0 || levelId >= 32)
  {
    gameData->parseError = "Invalid level id";
    return 0;
  }
  
  gameData->lastLevelIndex = levelId;
  Level *level = &gameData->levels[levelId];

  while (1)
  {
    // try to read the next token and if we don't know how to handle it,
    // we step back and return
    ParserState prevState = *state;
    
    if (!ParserStateReadNextToken(state))
    {
      // end of file
      return 1;
    }

    if (TextIsEqual(state->nextToken, "initialGold"))
    {
      if (!ParserStateReadNextInt(state, &level->initialGold))
      {
        gameData->parseError = "Failed to read initial gold";
        return 0;
      }

      continue;
    }

    // no match, return to previous state and break
    *state = prevState;
    break;
  }

  printf("Parsed level %d, initialGold=%d\n", levelId, level->initialGold);
  return 1;
}

int ParseGameData(ParserState *state, ParsedGameData *gameData)
{
  *gameData = (ParsedGameData){0};
  gameData->lastLevelIndex = -1;

  while (ParserStateReadNextToken(state))
  {
    if (TextIsEqual(state->nextToken, "Level"))
    {
      // read id of level
      if (ParseGameDataLevelSection(state, gameData) == 0)
      {
        return 0;
      }
    }


  }

  return 1;
}

int main()
{
  char *fileContent = LoadFileText("data/level.txt");
  if (fileContent == NULL)
  {
    printf("Failed to load file\n");
    return 1;
  }

  ParserState state = {fileContent, 0, {0}};
  ParsedGameData gameData = {0};

  if (!ParseGameData(&state, &gameData))
  {
    printf("Failed to parse game data: %s\n", gameData.parseError);
    UnloadFileText(fileContent);
    return 1;
  }

  UnloadFileText(fileContent);
  return 0; 
}
